@@ -1,33 +1,19 @@
 /*
- * Copyright (c) 2000-2002 Silicon Graphics, Inc.  All Rights Reserved.
+ * Copyright (c) 2000-2002 Silicon Graphics, Inc.
+ * All Rights Reserved.
  *
- * This program is free software; you can redistribute it and/or modify it
- * under the terms of version 2 of the GNU General Public License as
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation.
  *
- * This program is distributed in the hope that it would be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+ * This program is distributed in the hope that it would be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
  *
- * Further, this software is distributed without any warranty that it is
- * free of the rightful claim of any third person regarding infringement
- * or the like.  Any license provided herein, whether implied or
- * otherwise, applies only to this software file.  Patent licenses, if
- * any, provided herein do not apply to combinations of this program with
- * other software, or any other product whatsoever.
- *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write the Free Software Foundation, Inc., 59
- * Temple Place - Suite 330, Boston MA 02111-1307, USA.
- *
- * Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
- * Mountain View, CA  94043, or:
- *
- * http://www.sgi.com
- *
- * For further information regarding this notice, see:
- *
- * http://oss.sgi.com/projects/GenInfo/SGIGPLNoticeExplan/
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write the Free Software Foundation,
+ * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /*
@@ -35,7 +21,7 @@
  *
  * fsr [-d] [-v] [-n] [-s] [-g] [-t mins] [-f leftf] [-m mtab]
  * fsr [-d] [-v] [-n] [-s] [-g] xfsdev | dir | file ...
- * 
+ *
  * If invoked in the first form fsr does the following: starting with the
  * dev/inum specified in /etc/fsrlast this reorgs each reg file in each file
  * system found in /etc/mtab.  After 2 hours of this we record the current
@@ -55,7 +41,7 @@
  *	-p passes	Number of passes before terminating global re-org.
  */
 
-#include <xfs/libxfs.h>
+#include <xfs/xfs.h>
 #include <xfs/jdm.h>
 #include "config.h"
 
@@ -72,6 +58,12 @@
 #include <attr/attributes.h>
 #include <xfs/xfs_dfrag.h>
 
+
+#ifndef XFS_XFLAG_NODEFRAG
+#define XFS_XFLAG_NODEFRAG 0x00002000 /* src dependancy, remove later */
+#endif
+
+
 char *progname;
 
 int vflag;
@@ -115,10 +107,10 @@ static int	pagesize;
 
 void usage(int ret);
 static int  fsrfile(char *fname, xfs_ino_t ino);
-static int  fsrfile_common( char *fname, char *tname, char *mnt, 
+static int  fsrfile_common( char *fname, char *tname, char *mnt,
                             int fd, xfs_bstat_t *statp);
-static int  packfile(char *fname, char *tname, int fd, 
-                     xfs_bstat_t *statp, int flag);
+static int  packfile(char *fname, char *tname, int fd,
+                     xfs_bstat_t *statp, struct fsxattr *fsxp);
 static void fsrdir(char *dirname);
 static int  fsrfs(char *mntdir, xfs_ino_t ino, int targetrange);
 static void initallfs(char *mtab);
@@ -136,7 +128,7 @@ static void tmp_init(char *mnt);
 static char * tmp_next(char *mnt);
 static void tmp_close(char *mnt);
 int xfs_getgeom(int , xfs_fsop_geom_v1_t * );
-static int getmntany (FILE *, struct mntent *, struct mntent *);
+static int getmntany(FILE *, struct mntent *, struct mntent *, struct stat64 *);
 
 xfs_fsop_geom_v1_t fsgeom;	/* geometry of active mounted system */
 
@@ -155,17 +147,17 @@ int		nfrags = 0;	/* Debug option: Coerse into specific number
 				 * of extents */
 int		openopts = O_CREAT|O_EXCL|O_RDWR|O_DIRECT;
 
-int 
+int
 xfs_fsgeometry(int fd, xfs_fsop_geom_v1_t *geom)
 {
     return ioctl(fd, XFS_IOC_FSGEOMETRY_V1, geom);
 }
 
-int 
+int
 xfs_bulkstat_single(int fd, xfs_ino_t *lastip, xfs_bstat_t *ubuffer)
 {
     xfs_fsop_bulkreq_t  bulkreq;
-    
+
     bulkreq.lastip = lastip;
     bulkreq.icount = 1;
     bulkreq.ubuffer = ubuffer;
@@ -173,12 +165,12 @@ xfs_bulkstat_single(int fd, xfs_ino_t *lastip, xfs_bstat_t *ubuffer)
     return ioctl(fd, XFS_IOC_FSBULKSTAT_SINGLE, &bulkreq);
 }
 
-int 
-xfs_bulkstat(int fd, xfs_ino_t *lastip, int icount, 
+int
+xfs_bulkstat(int fd, xfs_ino_t *lastip, int icount,
                     xfs_bstat_t *ubuffer, __s32 *ocount)
 {
     xfs_fsop_bulkreq_t  bulkreq;
-    
+
     bulkreq.lastip = lastip;
     bulkreq.icount = icount;
     bulkreq.ubuffer = ubuffer;
@@ -186,13 +178,13 @@ xfs_bulkstat(int fd, xfs_ino_t *lastip, int icount,
     return ioctl(fd, XFS_IOC_FSBULKSTAT, &bulkreq);
 }
 
-int 
+int
 xfs_swapext(int fd, xfs_swapext_t *sx)
 {
     return ioctl(fd, XFS_IOC_SWAPEXT, sx);
 }
 
-int 
+int
 xfs_fscounts(int fd, xfs_fsop_counts_t *counts)
 {
     return ioctl(fd, XFS_IOC_FSCOUNTS, counts);
@@ -208,13 +200,13 @@ aborter(int unused)
 int
 main(int argc, char **argv)
 {
-	struct stat sb, sb2;
+	struct stat64 sb, sb2;
 	char *argname;
 	char *cp;
 	int c;
 	struct mntent mntpref;
 	register struct mntent *mntp;
-	struct mntent mntent;
+	struct mntent ment;
 	register FILE *mtabp;
 
 	setlinebuf(stdout);
@@ -285,25 +277,25 @@ main(int argc, char **argv)
 
 	/* Save the caller's real uid */
 	RealUid = getuid();
-	
+
 	pagesize = getpagesize();
 
 	if (optind < argc) {
 		for (; optind < argc; optind++) {
 			argname = argv[optind];
 			mntp = NULL;
-			if (lstat(argname, &sb) < 0) {
+			if (lstat64(argname, &sb) < 0) {
 				fprintf(stderr,
 					_("%s: could not stat: %s: %s\n"),
 					progname, argname, strerror(errno));
 				continue;
 			}
-			if (S_ISLNK(sb.st_mode) && stat(argname, &sb2) == 0 &&
+			if (S_ISLNK(sb.st_mode) && stat64(argname, &sb2) == 0 &&
 			    (S_ISBLK(sb2.st_mode) || S_ISCHR(sb2.st_mode)))
 				sb = sb2;
 			if (S_ISBLK(sb.st_mode) || (S_ISDIR(sb.st_mode))) {
 				if ((mtabp = setmntent(mtab, "r")) == NULL) {
-					fprintf(stderr, 
+					fprintf(stderr,
 						_("%s: cannot read %s\n"),
 						progname, mtab);
 					exit(1);
@@ -314,18 +306,16 @@ main(int argc, char **argv)
 				else
 					mntpref.mnt_fsname = argname;
 
-				if ((getmntany(mtabp, &mntent, &mntpref) == 0) 
-				     &&
-				   (strcmp(mntent.mnt_type, MNTTYPE_XFS) == 0))
-				{
-					mntp = &mntent;
+				if (getmntany(mtabp, &ment, &mntpref, &sb) &&
+				    strcmp(ment.mnt_type, MNTTYPE_XFS) == 0) {
+					mntp = &ment;
 					if (S_ISBLK(sb.st_mode)) {
 						cp = mntp->mnt_dir;
-						if (cp == NULL || 
-						    stat(cp, &sb2) < 0) {
+						if (cp == NULL ||
+						    stat64(cp, &sb2) < 0) {
 							fprintf(stderr, _(
 						"%s: could not stat: %s: %s\n"),
-							progname, argname, 
+							progname, argname,
 							strerror(errno));
 							continue;
 						}
@@ -342,10 +332,7 @@ main(int argc, char **argv)
 				        progname, argname);
 				exit(1);
 			} else if (S_ISDIR(sb.st_mode) || S_ISREG(sb.st_mode)) {
-				struct statfs fs;
-
-				statfs(argname, &fs);
-				if (fs.f_type != XFS_SB_MAGIC) {
+				if (!platform_test_xfs_path(argname)) {
 					fprintf(stderr, _(
 				        "%s: cannot defragment: %s: Not XFS\n"),
 				        progname, argname);
@@ -369,7 +356,7 @@ main(int argc, char **argv)
 }
 
 void
-usage(int ret) 
+usage(int ret)
 {
 	fprintf(stderr, _("Usage: %s [xfsfile] ...\n"), progname);
 	exit(ret);
@@ -385,7 +372,7 @@ initallfs(char *mtab)
 	struct mntent *mp;
 	int mi;
 	char *cp;
-	struct stat sb;
+	struct stat64 sb;
 
 	fp = setmntent(mtab, "r");
 	if (fp == NULL) {
@@ -394,12 +381,12 @@ initallfs(char *mtab)
 	}
 
 	/* malloc a number of descriptors, increased later if needed */
-	if ((fsbase = (fsdesc_t *)malloc(fsbufsize * sizeof(fsdesc_t))) == NULL) {
+	if (!(fsbase = (fsdesc_t *)malloc(fsbufsize * sizeof(fsdesc_t)))) {
 		fsrprintf(_("out of memory: %s\n"), strerror(errno));
 		exit(1);
 	}
 	fsend = (fsbase + fsbufsize - 1);
-	
+
 	/* find all rw xfs file systems */
 	mi = 0;
 	fs = fsbase;
@@ -407,7 +394,7 @@ initallfs(char *mtab)
 		int rw = 0;
 
 		if (strcmp(mp->mnt_type, MNTTYPE_XFS ) != 0 ||
-		    stat(mp->mnt_fsname, &sb) == -1 ||
+		    stat64(mp->mnt_fsname, &sb) == -1 ||
 		    !S_ISBLK(sb.st_mode))
 			continue;
 
@@ -425,7 +412,7 @@ initallfs(char *mtab)
 
 		if (mi == fsbufsize) {
 			fsbufsize += NMOUNT;
-			if ((fsbase = (fsdesc_t *)realloc((char *)fsbase, 
+			if ((fsbase = (fsdesc_t *)realloc((char *)fsbase,
 			              fsbufsize * sizeof(fsdesc_t))) == NULL) {
 				fsrprintf(_("out of memory: %s\n"),
 					strerror(errno));
@@ -465,7 +452,7 @@ initallfs(char *mtab)
 	}
 }
 
-static void 
+static void
 fsrallfs(int howlong, char *leftofffile)
 {
 	int fd;
@@ -478,7 +465,7 @@ fsrallfs(int howlong, char *leftofffile)
 	xfs_ino_t startino = 0;
 	fsdesc_t *fsp;
 	struct stat64 sb, sb2;
-	
+
 	fsrprintf("xfs_fsr -m %s -t %d -f %s ...\n", mtab, howlong, leftofffile);
 
 	endtime = starttime + howlong;
@@ -504,12 +491,12 @@ fsrallfs(int howlong, char *leftofffile)
 			{
 				fsrprintf(_("Can't use %s: mode=0%o own=%d"
 					" nlink=%d\n"),
-					leftofffile, sb.st_mode, 
+					leftofffile, sb.st_mode,
 					sb.st_uid, sb.st_nlink);
 				close(fd);
 				fd = NULLFD;
 			}
-		} 
+		}
 		else {
 			close(fd);
 			fd = NULLFD;
@@ -518,7 +505,7 @@ fsrallfs(int howlong, char *leftofffile)
 	else {
 		fd = NULLFD;
 	}
-			
+
 	if (fd != NULLFD) {
 		if (read(fd, buf, SMBUFSZ) == -1) {
 			fs = fsbase;
@@ -541,9 +528,7 @@ fsrallfs(int howlong, char *leftofffile)
 				startpass = atoi(++ptr);
 				ptr = strchr(ptr, ' ');
 				if (ptr) {
-					startino = strtoull(++ptr, 
-					                   (char **)NULL, 
-					                   10);
+					startino = strtoull(++ptr, NULL, 10);
 				}
 			}
 			if (startpass < 0)
@@ -629,7 +614,7 @@ fsrall_cleanup(int timeout)
 		          leftofffile, strerror(errno));
 	else {
 		if (timeout) {
-			ret = sprintf(buf, "%s %d %llu\n", fs->dev, 
+			ret = sprintf(buf, "%s %d %llu\n", fs->dev,
 			        fs->npass, (unsigned long long)leftoffino);
 			if (write(fd, buf, ret) < strlen(buf))
 				fsrprintf(_("write(%s) failed: %s\n"),
@@ -685,11 +670,12 @@ fsrfs(char *mntdir, xfs_ino_t startino, int targetrange)
 
 	tmp_init(mntdir);
 
-	while (! xfs_bulkstat(fsfd, &lastino, GRABSZ, &buf[0], &buflenout)) {
+	while ((ret = xfs_bulkstat(fsfd,
+				&lastino, GRABSZ, &buf[0], &buflenout) == 0)) {
 		xfs_bstat_t *p;
 		xfs_bstat_t *endp;
 
-		if (buflenout == 0 )
+		if (buflenout == 0)
 			goto out0;
 
 		/* Each loop through, defrag targetrange percent of the files */
@@ -697,13 +683,13 @@ fsrfs(char *mntdir, xfs_ino_t startino, int targetrange)
 
 		qsort((char *)buf, buflenout, sizeof(struct xfs_bstat), cmp);
 
-		for ( p = buf, endp = (buf + buflenout); p < endp ; p++ ) {
+		for (p = buf, endp = (buf + buflenout); p < endp ; p++) {
 			/* Do some obvious checks now */
 			if (((p->bs_mode & S_IFMT) != S_IFREG) ||
-			     (p->bs_extents < 2)) 
+			     (p->bs_extents < 2))
 				continue;
 
-			if((fd = jdm_open(fshandlep, p, O_RDWR)) < 0) {
+			if ((fd = jdm_open(fshandlep, p, O_RDWR)) < 0) {
 				/* This probably means the file was
 				 * removed while in progress of handling
 				 * it.  Just quietly ignore this file.
@@ -715,13 +701,13 @@ fsrfs(char *mntdir, xfs_ino_t startino, int targetrange)
 			}
 
 			/* Don't know the pathname, so make up something */
-			sprintf(fname, "ino=%llu", (unsigned long long)p->bs_ino);
+			sprintf(fname, "ino=%lld", (long long)p->bs_ino);
 
 			/* Get a tmp file name */
 			tname = tmp_next(mntdir);
 
 			ret = fsrfile_common(fname, tname, mntdir, fd, p);
-			
+
 			leftoffino = p->bs_ino;
 
 			close(fd);
@@ -738,12 +724,14 @@ fsrfs(char *mntdir, xfs_ino_t startino, int targetrange)
 			exit(1);
 		}
 	}
+	if (ret < 0)
+		fsrprintf(_("%s: xfs_bulkstat: %s\n"), progname, strerror(errno));
 out0:
 	tmp_close(mntdir);
 	close(fsfd);
 	return 0;
 }
-		
+
 /*
  * To compare bstat structs for qsort.
  */
@@ -767,8 +755,8 @@ fsrdir(char *dirname)
 }
 
 /*
- * Sets up the defragmentation of a file based on the 
- * filepath.  It collects the bstat information, does 
+ * Sets up the defragmentation of a file based on the
+ * filepath.  It collects the bstat information, does
  * an open on the file and passes this all to fsrfile_common.
  */
 static int
@@ -797,14 +785,14 @@ fsrfile(char *fname, xfs_ino_t ino)
 			fname, strerror(errno));
 		return -1;
 	}
-	
+
 	if ((xfs_bulkstat_single(fsfd, &ino, &statbuf)) < 0) {
 		fsrprintf(_("unable to get bstat on %s: %s\n"),
 			fname, strerror(errno));
 		close(fsfd);
 		return -1;
 	}
-		
+
 	fd = jdm_open( fshandlep, &statbuf, O_RDWR);
 	if (fd < 0) {
 		fsrprintf(_("unable to open handle %s: %s\n"),
@@ -836,21 +824,21 @@ fsrfile(char *fname, xfs_ino_t ino)
 /*
  * This is the common defrag code for either a full fs
  * defragmentation or a single file.  Check as much as
- * possible with the file, fork a process to setuid to the 
+ * possible with the file, fork a process to setuid to the
  * target file owner's uid and defragment the file.
- * This is done so the new extents created in a tmp file are 
- * reflected in the owners' quota without having to do any 
- * special code in the kernel.  When the existing extents 
- * are removed, the quotas will be correct.  It's ugly but 
- * it saves us from doing some quota  re-construction in 
+ * This is done so the new extents created in a tmp file are
+ * reflected in the owners' quota without having to do any
+ * special code in the kernel.  When the existing extents
+ * are removed, the quotas will be correct.  It's ugly but
+ * it saves us from doing some quota  re-construction in
  * the extent swap.  The price is that the defragmentation
  * will fail if the owner of the target file is already at
  * their quota limit.
  */
 static int
 fsrfile_common(
-	char		*fname, 
-	char		*tname, 
+	char		*fname,
+	char		*tname,
 	char		*fsname,
 	int		fd,
 	xfs_bstat_t	*statp)
@@ -858,9 +846,8 @@ fsrfile_common(
 	int		error;
 	struct statvfs64 vfss;
 	struct fsxattr	fsx;
-	int		do_rt = 0;
 	unsigned long	bsize;
-		
+
 	if (vflag)
 		fsrprintf("%s\n", fname);
 
@@ -875,9 +862,9 @@ fsrfile_common(
 		return(0);
 	}
 
-	/* Check if a mandatory lock is set on the file to try and 
+	/* Check if a mandatory lock is set on the file to try and
 	 * avoid blocking indefinitely on the reads later. Note that
-	 * someone could still set a mandatory lock after this check 
+	 * someone could still set a mandatory lock after this check
 	 * but before all reads have completed to block fsr reads.
 	 * This change just closes the window a bit.
 	 */
@@ -917,12 +904,22 @@ fsrfile_common(
 		return 1;
 	}
 
-	/* Check realtime info */
 	if ((ioctl(fd, XFS_IOC_FSGETXATTR, &fsx)) < 0) {
-		fsrprintf(_("failed to get attrs: %s\n"), fname);
+		fsrprintf(_("failed to get inode attrs: %s\n"), fname);
 		return(-1);
 	}
-	if (fsx.fsx_xflags == XFS_XFLAG_REALTIME) {
+	if (fsx.fsx_xflags & (XFS_XFLAG_IMMUTABLE|XFS_XFLAG_APPEND)) {
+		if (vflag)
+			fsrprintf(_("%s: immutable/append, ignoring\n"), fname);
+		return(0);
+	}
+	if (fsx.fsx_xflags & XFS_XFLAG_NODEFRAG) {
+		if (vflag)
+			fsrprintf(_("%s: marked as don't defrag, ignoring\n"), 
+			    fname);
+		return(0);
+	}
+	if (fsx.fsx_xflags & XFS_XFLAG_REALTIME) {
 		if (xfs_getrt(fd, &vfss) < 0) {
 			fsrprintf(_("cannot get realtime geometry for: %s\n"),
 				fname);
@@ -933,7 +930,6 @@ fsrfile_common(
 				"ignoring file\n"), fname);
 			return(-1);
 		}
-		do_rt = 1;
 	}
 
 	if ((RealUid != ROOT) && (RealUid != statp->bs_uid)) {
@@ -941,7 +937,7 @@ fsrfile_common(
 		return -1;
 	}
 
-	/* 
+	/*
 	 * Previously the code forked here, & the child changed it's uid to
 	 * that of the file's owner and then called packfile(), to keep
 	 * quota counts correct.  (defragged files could use fewer blocks).
@@ -950,7 +946,7 @@ fsrfile_common(
 	 * file we're defragging, in packfile().
 	 */
 
-	if ((error = packfile(fname, tname, fd, statp, do_rt)))
+	if ((error = packfile(fname, tname, fd, statp, &fsx)))
 		return error;
 	return -1; /* no error */
 }
@@ -958,13 +954,14 @@ fsrfile_common(
 
 /*
  * Do the defragmentation of a single file.
- * We already are pretty sure we can and want to 
+ * We already are pretty sure we can and want to
  * defragment the file.  Create the tmp file, copy
  * the data (maintaining holes) and call the kernel
  * extent swap routinte.
  */
 static int
-packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
+packfile(char *fname, char *tname, int fd,
+	 xfs_bstat_t *statp, struct fsxattr *fsxp)
 {
 	int 		tfd;
 	int		srval;
@@ -977,19 +974,18 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 	off64_t 	cnt, pos;
 	void 		*fbuf;
 	int 		ct, wc, wc_b4;
-	struct fsxattr  tfsx;
 	char		ffname[SMBUFSZ];
 	int		ffd = -1;
 
 	/*
 	 * Work out the extent map - nextents will be set to the
-	 * minimum number of extents needed for the file (taking 
+	 * minimum number of extents needed for the file (taking
 	 * into account holes), cur_nextents is the current number
 	 * of extents.
 	 */
 	nextents = read_fd_bmap(fd, statp, &cur_nextents);
 
-	if ( cur_nextents == 1 || cur_nextents <= nextents ) {
+	if (cur_nextents == 1 || cur_nextents <= nextents) {
 		if (vflag)
 			fsrprintf(_("%s already fully defragmented.\n"), fname);
 		return 1; /* indicates no change/no error */
@@ -997,19 +993,19 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 
 	if (dflag)
 		fsrprintf(_("%s extents=%d can_save=%d tmp=%s\n"),
-		          fname, cur_nextents, (cur_nextents - nextents), 
+		          fname, cur_nextents, (cur_nextents - nextents),
 		          tname);
 
 	if ((tfd = open(tname, openopts, 0666)) < 0) {
 		if (vflag)
-			fsrprintf(_("could not open tmp as uid %d: %s: %s\n"),
-				   statp->bs_uid,tname, strerror(errno));
+			fsrprintf(_("could not open tmp file: %s: %s\n"),
+				   tname, strerror(errno));
 		return -1;
 	}
 	unlink(tname);
 
 	/* Setup extended attributes */
-	if( statp->bs_xflags & XFS_XFLAG_HASATTR ) {
+	if (statp->bs_xflags & XFS_XFLAG_HASATTR) {
 		if (attr_setf(tfd, "X", "X", 1, ATTR_CREATE) != 0) {
 			fsrprintf(_("could not set ATTR on tmp: %s:\n"), tname);
 			close(tfd);
@@ -1019,35 +1015,26 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 			fsrprintf(_("%s set temp attr\n"), tname);
 	}
 
-	if ((ioctl(tfd, XFS_IOC_DIOINFO, &dio)) < 0 ) {
-		fsrprintf(_("could not get I/O info on tmp: %s\n"), tname);
-		close(tfd);
-		return -1;
-	}
-	if (do_rt) {
-		int rt_textsize = fsgeom.rtextsize * fsgeom.blocksize;
-
-		tfsx.fsx_xflags = XFS_XFLAG_REALTIME;
-
-		if ((tfsx.fsx_extsize = rt_textsize) <= 0 ) {
-			fsrprintf(_("realtime geometry not available for "
-				"tmp: %s\n"), fname);
-			close(tfd);
-			return -1;
-		}
-
-		if (ioctl( tfd,  XFS_IOC_FSSETXATTR, &tfsx) < 0) {
-			fsrprintf(_("could not set realtime flag on tmp: %s\n"),
+	/* Setup extended inode flags, project identifier, etc */
+	if (fsxp->fsx_xflags || fsxp->fsx_projid) {
+		if (ioctl(tfd, XFS_IOC_FSSETXATTR, fsxp) < 0) {
+			fsrprintf(_("could not set inode attrs on tmp: %s\n"),
 				tname);
 			close(tfd);
 			return -1;
 		}
 	}
 
-	dio_min   = dio.d_miniosz;
-	if (statp->bs_size <= dio_min)
+	if ((ioctl(tfd, XFS_IOC_DIOINFO, &dio)) < 0 ) {
+		fsrprintf(_("could not get DirectIO info on tmp: %s\n"), tname);
+		close(tfd);
+		return -1;
+	}
+
+	dio_min = dio.d_miniosz;
+	if (statp->bs_size <= dio_min) {
 		blksz_dio = dio_min;
-	else {
+	} else {
 		blksz_dio = min(dio.d_maxiosz, BUFFER_MAX - pagesize);
 		if (argv_blksz_dio != 0)
 			blksz_dio = min(argv_blksz_dio, blksz_dio);
@@ -1061,8 +1048,7 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 			dio.d_maxiosz, pagesize);
 	}
 
-	/* Malloc a buffer */
-	if (! (fbuf = (char *)memalign(dio.d_mem, blksz_dio))) {
+	if (!(fbuf = (char *)memalign(dio.d_mem, blksz_dio))) {
 		fsrprintf(_("could not allocate buf: %s\n"), tname);
 		close(tfd);
 		return -1;
@@ -1076,6 +1062,8 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 		if ((ffd = open(ffname, openopts, 0666)) < 0) {
 			fsrprintf(_("could not open fragfile: %s : %s\n"),
 				   ffname, strerror(errno));
+			close(tfd);
+			free(fbuf);
 			return -1;
 		}
 		unlink(ffname);
@@ -1128,6 +1116,7 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 	if (cur_nextents <= new_nextents) {
 		if (vflag)
 			fsrprintf(_("No improvement will be made (skipping): %s\n"), fname);
+		free(fbuf);
 		close(tfd);
 		return 1; /* no change/no error */
 	}
@@ -1150,7 +1139,7 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 			} else if (cnt % dio_min == 0) {
 				ct = min(cnt, blksz_dio);
 			} else {
-				ct = min(cnt + dio_min - (cnt % dio_min), 
+				ct = min(cnt + dio_min - (cnt % dio_min),
 					blksz_dio);
 			}
 			ct = read(fd, fbuf, ct);
@@ -1201,6 +1190,7 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 					}
 				}
 				free(fbuf);
+				close(tfd);
 				return -1;
 			}
 			if (nfrags) {
@@ -1240,7 +1230,7 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 	srval = xfs_swapext(fd, &sx);
 	if (srval < 0) {
 		if (errno == ENOTSUP) {
-			if (vflag || dflag) 
+			if (vflag || dflag)
 			   fsrprintf(_("%s: file type not supported\n"), fname);
 		} else if (errno == EFAULT) {
 			/* The file has changed since we started the copy */
@@ -1262,7 +1252,7 @@ packfile(char *fname, char *tname, int fd, xfs_bstat_t *statp, int do_rt)
 	/* Report progress */
 	if (vflag)
 		fsrprintf(_("extents before:%d after:%d %s %s\n"),
-			  cur_nextents, new_nextents, 
+			  cur_nextents, new_nextents,
 			  (new_nextents <= nextents ? "DONE" : "    " ),
 		          fname);
 	close(tfd);
@@ -1413,7 +1403,7 @@ int	read_fd_bmap(int fd, xfs_bstat_t *sin, int *cur_nextents)
 /*
  * Read the block map and return the number of extents.
  */
-int 
+int
 getnextents(int fd)
 {
 	int		nextents;
@@ -1452,7 +1442,7 @@ xfs_getgeom(int fd, xfs_fsop_geom_v1_t * fsgeom)
 }
 
 /*
- * Get xfs realtime space information 
+ * Get xfs realtime space information
  */
 int
 xfs_getrt(int fd, struct statvfs64 *sfbp)
@@ -1497,20 +1487,28 @@ fsrprintf(const char *fmt, ...)
  * emulate getmntany
  */
 static int
-getmntany (FILE *filep, struct mntent *mp, struct mntent *mpref)
+getmntany(FILE *fp, struct mntent *mp, struct mntent *mpref, struct stat64 *s)
 {
-        int match = 0;
-        struct mntent *t = NULL;
+	struct mntent *t;
+	struct stat64 ms;
 
-        while (!match && (t = getmntent(filep)) != 0) {
-                if (mpref->mnt_fsname != NULL &&
-                  strcmp(mpref->mnt_fsname, t->mnt_fsname) != 0)  continue;
-                if (mpref->mnt_dir != NULL &&
-                    strcmp(mpref->mnt_dir, t->mnt_dir) != 0)  continue;
-                match++;
-        }
-        if (match) *mp = *t;
-        return !match;
+	while ((t = getmntent(fp))) {
+		if (mpref->mnt_fsname) {	/* device */
+			if (stat64(t->mnt_fsname, &ms) < 0)
+				continue;
+			if (s->st_rdev != ms.st_rdev)
+				continue;
+		}
+		if (mpref->mnt_dir) {		/* mount point */
+			if (stat64(t->mnt_dir, &ms) < 0)
+				continue;
+			if (s->st_ino != ms.st_ino || s->st_dev != ms.st_dev)
+				continue;
+		}
+		*mp = *t;
+		break;
+	}
+	return (t != NULL);
 }
 
 
@@ -1569,9 +1567,9 @@ tmp_next(char *mnt)
 {
 	static char	buf[SMBUFSZ];
 
-	sprintf(buf, "%s/.fsr/ag%d/tmp%d", 
-	        ( (strcmp(mnt, "/") == 0) ? "" : mnt), 
-	        tmp_agi, 
+	sprintf(buf, "%s/.fsr/ag%d/tmp%d",
+	        ( (strcmp(mnt, "/") == 0) ? "" : mnt),
+	        tmp_agi,
 	        getpid());
 
 	if (++tmp_agi == fsgeom.agcount)
